from distutils.version import StrictVersion

from psij.descriptor import Descriptor


__PSI_J_EXECUTORS__ = [Descriptor(name='rest', version=StrictVersion('0.0.1'),
                                  cls='psij.executors.rest_service.RestServiceJobExecutor')]



"""This module contains the RestService :class:`~psij.JobExecutor`."""

import json
import requests
import websocket
import threading

from urllib.parse import urlparse, urlunparse

from typing import Optional, List, Dict

from psij import (
    Job,
    JobExecutorConfig,
    JobState,
    JobStatus,
    JobExecutor,
    Export,
)


class RestServiceJobExecutor(JobExecutor):
    """A :class:`~psij.JobExecutor` for a PSI/J Rest service endpoint.

    This executor forwards all requests to a Rest service endpoint which is then
    executing the respective request on the target resource.
    """

    _state_map = {'NEW': JobState.NEW,
                  'QUEUED': JobState.QUEUED,
                  'ACTIVE': JobState.ACTIVE,
                  'COMPLETED': JobState.COMPLETED,
                  'FAILED': JobState.FAILED,
                  'CANCELED': JobState.CANCELED}
    _final = [JobState.COMPLETED, JobState.FAILED, JobState.CANCELED]

    def __init__(self, url: Optional[str] = None,
                 config: Optional[JobExecutorConfig] = None) -> None:
        """
        Initializes a `RestServiceJobExecutor`.

        :param url: address at which to contact the remote service.
                    Supported schemas: `http://` and `https://`
        :param config: The `RestServiceJobExecutor` does not have any
                    configuration options.
        """
        if not config:
            config = JobExecutorConfig()

        if not url:
            raise ValueError('REST executor requires service URL')

        url_obj = urlparse(url)

        if url_obj.scheme not in ['http', 'https']:
            raise ValueError('expected `http://` or `https://` as url schema')

        if url_obj.path and url_obj.path != '/':
            name = url_obj.path.strip('/')
        else:
            name = 'local'

        url = urlunparse(url_obj._replace(path=''))

        super().__init__(url=str(url), config=config)

        self._jobs: Dict[str, Job] = dict()
        self._idmap: Dict[str, str] = dict()
        self._serialize = Export()

        # we can only update the idmap when the `submit` request returns, but
        # an state notification may happen before that.  We use a lock to ensure
        # that state updates are delayed until after `submit` completed.
        self._lock = threading.Lock()

        # connect to service and register this client instance
        self._cid = str(requests.get('%s/executor/%s' % (url, name)).json())

        # create a daemong thread for websocket state notifications
        t = threading.Thread(target=self._state_listener)
        t.daemon = True
        t.start()

    def _state_listener(self) -> None:
        """thrtead to listen for on job state update messages on the websocket.

        Update the status of the psij.Job.
        """
        if 'http://' in self.url:
            ws_url = self.url.replace('http://', 'ws://')
        else:
            ws_url = self.url.replace('http://', 'ws://')

        assert ws_url.startswith('ws://')

        ws = websocket.create_connection(ws_url + '/ws/' + self._cid)
        while True:
            msg = json.loads(ws.recv())

            with self._lock:
                jobid = self._idmap.get(msg['jobid'])
                if not jobid:
                    # FIXME: use logger
                    print('job %s unknown: %s' % (jobid, self._idmap.keys()))
                    return

            job = self._jobs.get(jobid)
            assert job

            state = self._state_map[msg['state']]
            status = JobStatus(state, time=msg['time'], message=msg['message'],
                               exit_code=msg['exit_code'], metadata=msg['metadata'])
            self._set_job_status(job, status)

            if state in self._final:
                del self._jobs[jobid]

    def submit(self, job: Job) -> None:
        """See :func:`~psij.job_executor.JobExecutor.submit`."""
        job.executor = self
        with self._lock:
            self._jobs[job.id] = job
            spec = self._serialize.to_dict(job.spec)
            rep = requests.put('%s/%s' % (self.url, self._cid), json=spec)
            job._native_id = str(rep.json())
            self._idmap[job._native_id] = job.id

    def cancel(self, job: Job) -> None:
        """See :func:`~psij.job_executor.JobExecutor.cancel`."""
        requests.delete('%s/%s/%s' % (self.url, self._cid, job._native_id))

    def list(self) -> List[str]:
        """See :func:`~psij.job_executor.JobExecutor.list`.

        Return a list of ids representing jobs that are running on the
        underlying implementation.  We consider the remote service's job ids as
        native job ids and return them unaltered.

        :return: The list of known job ids.
        """
        rep = requests.get('%s/%s/jobs' % (self.url, self._cid))
        return rep.json()

    def attach(self, job: Job, native_id: str) -> None:
        """
        Attaches a job instance to an existing job.

        The job instance must be in the :attr:`~psij.JobState.NEW` state.

        :param job: The job instance to attach.
        :param native_id: The native ID of the backend job to attached to, as
          obtained through the `:func:list` method.
        """
        assert job.status.state == JobState.NEW
        job.executor = self
        job._native_id = native_id
        self._idmap[native_id] = job.id
        self._jobs[job.id] = job



#!/usr/bin/env python3

'''
This file implements a psij job service.  The service can be contacted via
a REST API.

the service will

The service interface mirrors that of an psij executor implementation.  The
supported `cmd` requests, their parameters and return values are as follows:

    register_client(name: str = None) -> str

        REST: GET /executor/{name}/
        name: name of backend psij executor to use for this client
              defaults to `local`

        returns: a unique ID to identify the client on further requests.

        Register client, configure the service's psij executor to use for this
        client, and return a unique client ID.  That ID is required for
        all further requests.


    submit(cid: str, spec: Dict[str, Any]) -> str

        REST: PUT /{cid}/
        cid : client ID obtained via `register_client`
        spec: serialized psij.JobSepc

        This method submits a job as described by the JobSpec to the backend psij
        executor and returns the job ID.


    cancel(cid: str, jobid: str) -> None

        REST : DELETE /{cid}/{jobid}
        cid  : client ID obtained via `register_client`
        jobid: job ID obtained via `submit` or `list`

        This method will cancel the specified job.  The method returns without
        waiting for the cancelation request to suceed, the callee should observe
        state notifications to confirm successfull cancellation.


    list(cid: str) -> List[str]

        REST: GET /{cid}/jobs
        cid : client ID obtained via `register_client`

        This method will return a list of job IDs known to this service.

    FIXME:
      - use cookie instead of client id
      - add authorization and authentication
      - add data staging
'''


from fastapi import FastAPI, WebSocket, WebSocketDisconnect

import sys
import psij
import queue
import asyncio
import logging
import uvicorn
import functools

from typing import List, Dict, Any, Optional


# ------------------------------------------------------------------------------
#
class _Client(object):

    def __init__(self, jex: psij.JobExecutor):

        self.jex: psij.JobExecutor = jex
        self._jobs: Dict[str, psij.Job] = dict()
        self.ws: Optional[WebSocket] = None
        self._queue: queue.Queue = queue.Queue()

    def add_job(self, job: psij.Job) -> None:

        self._jobs[job.id] = job

    def get_job(self, jobid: str) -> Optional[psij.Job]:

        return self._jobs.get(jobid)

    def list_jobs(self) -> List[str]:

        return list(self._jobs.keys())

    def get_msg(self) -> Any:
        try:
            return self._queue.get_nowait()
        except Exception:
            pass

    def send(self, msg: Any) -> None:
        self._queue.put(msg)


# ------------------------------------------------------------------------------
#
class Service(object):

    def __init__(self, app: FastAPI) -> None:

        self._clients: Dict[str, _Client] = dict()
        self._deserialize = psij.Import()
        self._log = logging.getLogger('psij')
        self._cnt: int = 0

        # ----------------------------------------------------------------------
        # websocket endpoint at which client can regoister for state updates
        @app.websocket("/ws/{cid}")
        async def ws_endpoint(ws: WebSocket, cid: str) -> None:

            await ws.accept()

            client = self._clients.get(cid)
            if not client:
                self._log.error("refuse ws for %s" % cid)
                raise ValueError("unknown client cid %s" % cid)

            self._log.info("accept ws for %s" % cid)
            try:
                # keep this async task alive as long as there are messages to
                # send and the websocket is alive
                while True:
                    msg = client.get_msg()
                    if msg:
                        await ws.send_json(msg)
                    else:
                        await asyncio.sleep(0.1)

            except WebSocketDisconnect:
                self._log.info("dropped ws for %s" % cid)
        # ----------------------------------------------------------------------

    def _status_callback(self, cid: str, job: psij.Job, status: psij.JobStatus) -> None:

        client = self._clients.get(cid)
        if not client:
            print("unknown client cid %s" % cid)
            return

        msg = {'jobid': job.id,
               'time': status.time,
               'message': status.message,
               'state': str(status.state),
               'metadata': status.metadata,
               'exit_code': status.exit_code}
        self._log.debug('cb: %s: %s', cid, msg)
        client.send(msg)

    def _request_register(self, name: str, url: Optional[str] = None) -> str:
        '''
        parameters:
            name:str: name of psij executor to use
            url:str: optional URL to be passed to backend executor

        returns:
            str: unique ID identifying the registered client
        '''

        # register new client
        cid = 'client.%04d' % self._cnt
        self._cnt += 1

        # create executor
        jex = psij.JobExecutor.get_instance(name=name, url=url)

        # register state callback for this cid
        cb = functools.partial(self._status_callback, cid)
        jex.set_job_status_callback(cb)

        # store client information
        self._clients[cid] = _Client(jex)

        # client is now known and initialized
        return cid

    def _request_submit(self, cid: str, spec: Dict[str, Any]) -> str:
        '''
        parameters:
           cid:str   : client ID
           spec:Dict : job spec

        returns:
           jobid:str : psij job ID for submitted job
        '''

        client = self._clients.get(cid)
        if not client:
            raise ValueError('unknown client cid %s' % cid)

        job = psij.Job(self._deserialize.from_dict(spec, 'JobSpec'))
        client.add_job(job)
        client.jex.submit(job)

        return job.id

    def _request_cancel(self, cid: str, jobid: str) -> None:
        '''
        parameters:
           cid:str   : client ID
           jobid:str : psij job ID for job to be canceled
        '''

        client = self._clients.get(cid)
        if not client:
            raise ValueError('unknown client cid %s' % cid)

        job = client.get_job(jobid)
        if not job:
            raise ValueError('unknown job id %s' % jobid)

        client.jex.cancel(job)

    def _request_list(self, cid: str) -> List[str]:
        '''
        parameters:
           cid:str   : client ID

        returns:
           jobids:List[str] : all known psij job IDs
        '''

        client = self._clients.get(cid)
        if not client:
            raise ValueError('unknown client cid %s' % cid)

        return client.list_jobs()

    def _request_stage_in(self, data: str, fname: str) -> None:

        with open(fname, 'w') as fout:
            fout.write(data)

    def _request_stage_out(self, fname: str) -> str:

        with open(fname) as fin:
            return fin.read()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    app = FastAPI()
    service = Service(app)

    @app.get("/executor/{name}")
    def register(name: str, url: Optional[str] = None) -> str:
        """
        Register a new client and create an executor of type 'name'.
        request: GET /executor/{name}/
            name: type of executor to create for this client
            url: optional url to pass to the backend executor
        response: a new client ID (str)
        """

        return service._request_register(name, url)

    @app.put("/{cid}")
    def submit(cid: str, spec: Dict[str, Any]) -> str:
        """
        Submit a job.
        request: PUT /{cid}/
            cid: client ID as obtained by `register`
            data: json serialized `psij.JobSpec` dictionary
        response: a new job ID (str) for the submitted job
        """
        return service._request_submit(cid, spec)

    @app.delete("/{cid}/{jobid}")
    def cancel(cid: str, jobid: str) -> None:
        """
        Cancel a job.
        request: DELETE /{cid}/{jobid}
            cid: client ID as obtained by `register`
            jobid: ID of job to be canceled
        response: None
        """
        return service._request_cancel(cid, jobid)

    @app.get("/{cid}/jobs")
    def list_jobs(cid: str) -> List[str]:
        """
        List all known jobs.
        request: GET /{cid}/jobs
            cid: client ID as obtained by `register`
        response: serialized json string containing a list of job IDs (`List[str]`)
        """
        return service._request_list(cid)

    port = int(sys.argv[1])
    print('url: http://localhost:%d/' % port)
    uvicorn.run(app, port=port, access_log=False)



from setuptools import setup, find_packages

with open("src/psij-rest/VERSION") as f:
    VERSION = f.read().strip()

if __name__ == '__main__':
    with open('requirements.txt') as f:
        install_requires = f.readlines()

    setup(
        name='psij-rest',
        version=VERSION,

        description='''This is an implementation of the PSI/J (Portable Submission Interface for Jobs)
        specification.''',

        author='The ExaWorks Team',
        author_email='andre@merzky.net',

        url='https://github.com/exaworks/psij-rest',

        classifiers=[
            'Programming Language :: Python :: 3',
            'License :: OSI Approved :: MIT License',
        ],


        packages=find_packages(where='src') + ['psij-descriptors'],
        package_dir={'': 'src'},

        package_data={
            '': ['README.md', 'LICENSE']
        },

        scripts=['bin/psij_rest_service.py'],

        entry_points={
        },

        install_requires=install_requires,
        python_requires='>=3.7'
    )
